using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using HPS;

namespace hps_cs_samples
{
    public partial class Sample
    {
        public void Exchange_Component_Interaction(WindowKey wk, Canvas canvas, ref View view, Model model, SegmentKey modelKey)
        {
            try
            {
            /* Sample: Exchange_Component_Interaction
             * This sample will show you how to:
             * - Import and Exchange file and check for exceptions
             * - Traverse the component structure of a CADModel
             * - Obtain a component from a selection event
             * - Highlight a selected component
             * - Define named styles and glyphs
             * 
             * Read more about the topics covered in this sample in our Programming Guide 
             * - Integration with HOOPS Exchange
             * - Section 6.1:		Selection
             * - Section 6.2:		Highlighting
             * - Section 4.3:		Styles
             * - Section 4.5:		Glyphs */

            /* NOTE:
             * For this sample we are using an Action to traverse the components of the
             * Exchange file.
             * The same result could be obtain by putting the body of the action in a 
             * separate helper method */
            Action<Component, int, System.IO.StreamWriter> traverseComponents = null;
            traverseComponents = new Action<Component, int, System.IO.StreamWriter>((component, depth, writer) =>
                {
                    string outputString = "";

		            for(int i=0; i < depth; ++i)
			            outputString += "  ";
		
                    outputString += component.GetComponentType().ToString();

                    HPS.Metadata data = component.GetMetadata("Name");
                    HPS.StringMetadata metadata = new HPS.StringMetadata(data);
		            if(metadata.Type() != HPS.Type.None)
			            outputString += (" - Name: " + metadata.GetValue());
		            else
			            outputString += " - Name: N/A";

                    writer.WriteLine(outputString);

                    HPS.Component[] cArray = component.GetSubcomponents();

		            if(cArray.Length == 0)
			            return;
		            else
		            {
			            ++depth;
			            for(int i=0, e=cArray.Length; i < e; ++i)
			            {
				            traverseComponents(cArray[i], depth, writer);
			            }
		            }
                });

            /* This next block of code shows you how to load an Exchange file
             * 1. Create an ImportOptionsKit
             * 2. Use the Exchange::File::Import method to load the desired file, with the options you specified
             * 3. Since this operation happens on a separate thread, wait for it to complete before proceeding 
             * 4. Use a try-catch block to capture any Exchange IO exceptions 
             * 5. Obtain a CADModel object through the Exchange::ImportNotifier */
            HPS.Exchange.ImportNotifier exchangeNotifier = new HPS.Exchange.ImportNotifier();
            HPS.CADModel modelFile = new HPS.CADModel();
	        try {
		        string file_in = INPUT_FOLDER + "/revolved.prt";
                HPS.Exchange.ImportOptionsKit iok = new HPS.Exchange.ImportOptionsKit();
		        iok.SetBRepData(true);
                exchangeNotifier = HPS.Exchange.File.Import(file_in, iok);
		        exchangeNotifier.Wait();

		        modelFile = exchangeNotifier.GetCADModel();
		        view.Delete();
		        view = modelFile.ActivateDefaultCapture();
                canvas.AttachViewAsLayout(view);

                modelKey = view.GetAttachedModel().GetSegmentKey();
                modelKey.InsertDistantLight(new HPS.Vector(1, 1, 0));
		        view.FitWorld();
	        }

            catch (HPS.IOException)
            {
		        string error_message;
		        error_message = "Exchange Import Unsuccessful";
                modelKey.InsertText(new HPS.Point(0, 0, 0), error_message);
		        view.FitWorld();
		        view.Update();
	        }

            /* This block of code uses the lambda defined above to recursively walk the component
             * structure of the imported file, and output it to a text file.
             * 
             * For each component, its type and its name, when available, will be printed to the
             * output file.
             * 
             * The type of a component can be retrieved using the GetComponentType.
             * The name of a component is one of the metadata that can be associated with a 
             * component. To obtain it, simply call the GetMetadata method with "Name" as 
             * the argument.
             * 
             * If the type of the returned metadata is 'None' it means that the component
             * did not have a name metadata set on it */
            HPS.Component[] root = modelFile.GetSubcomponents();
            string file_out = OUTPUT_FOLDER + "/exchange_component_interaction.txt";
            using (System.IO.StreamWriter writer = new System.IO.StreamWriter(System.IO.File.OpenWrite(file_out)))
            {
                traverseComponents(root[0], 0, writer);
            }

            /* This block sets up the scene used in this sample.
	         * A segment is created which will be used to output text to the screen, and
	         * another segment is created to be used to house a marker which will indicate the
	         * location of the selection operation */
            HPS.SegmentKey elements = modelKey.Subsegment();
            elements.GetCameraControl().SetPosition(new HPS.Point(0, 0, 5)).
                SetTarget(new HPS.Point(0, 0, 0)).
                SetUpVector(new HPS.Vector(0, 1, 0)).
		        SetField(2, 2).
                SetProjection(HPS.Camera.Projection.Stretched);

            HPS.SegmentKey description = elements.Subsegment();
	        description.GetDrawingAttributeControl().SetDepthRange(0.0f, 0.1f);
	        description.GetVisibilityControl().SetEverything(false).SetText(true);

            description.GetTextAttributeControl().SetSize(0.02f, HPS.Text.SizeUnits.SubscreenRelative).
                SetAlignment(HPS.Text.Alignment.Center, HPS.Text.ReferenceFrame.WorldAligned, HPS.Text.Justification.CenterJustified);
            description.InsertText(new HPS.Point(0, -0.80f, 0), "A file describing the component structure of this model");
            description.InsertText(new HPS.Point(0, -0.85f, 0), "has been saved here: ");
            description.InsertText(new HPS.Point(0, -0.90f, 0), OUTPUT_FOLDER + "/exchange_component_interaction.txt");

	        /* This block of code defines the glyph symbol that will be used to draw the marker
	         * and the named style which will be used to highlight the selected component */
            HPS.PortfolioKey mpKey, vpKey;
	        modelKey.GetPortfolioControl().ShowTop(out mpKey);
            mpKey.DefineGlyph("myGlyph", HPS.GlyphKit.GetDefault(HPS.Glyph.Default.Plus));

	        view.GetSegmentKey().GetPortfolioControl().ShowTop(out vpKey);
            HPS.SegmentKey highlightStyle = view.GetSegmentKey().Subsegment();
	        highlightStyle.GetMaterialMappingControl().
                SetFaceColor(new HPS.RGBAColor(1, 0.5f, 0, 1)).
                SetEdgeColor(new HPS.RGBAColor(1, 0.5f, 0, 1));
	        vpKey.DefineNamedStyle("orangeHighlight", highlightStyle);

            HPS.MarkerKit markerKit = new HPS.MarkerKit();
            markerKit.SetPoint(new HPS.Point(-0.05f, -0.1f, 0));
            HPS.SegmentKey marker = elements.Subsegment();
	        marker.GetMarkerAttributeControl().SetSymbol("myGlyph").SetSize(3);
	        marker.InsertMarker(markerKit);
	        marker.GetDrawingAttributeControl().SetDepthRange(0.1f, 0.2f);
	        marker.GetVisibilityControl().SetMarkers(true);
            marker.GetMaterialMappingControl().SetMarkerColor(new HPS.RGBAColor(1, 0, 0, 1));
            marker.GetSelectabilityControl().SetEverything(HPS.Selectability.Value.Off);

	        /* This block of code does the following:
	         * 1. Sets up the selection options which will be used during the selection operation. In this case we decided
	         *		to select at the entity level, to sort the selection results from closer to most distant with respect 
	         *		to the camera, and to only consider the first selection element by setting the related limit to zero
	         * 2. Builds a key path. A key path is needed to correctly take into account cameras and matrix transforms 
	         *		when converting coordinates between systems. The key path is a collection of keys, going from the 
	         *		leaf to the root of the scene graph. The key path can omit some of the keys between the leaf and 
	         *		the root, but MUST contain all the include link between leaf and root 
	         * 3. Once the selection point has been properly converted to Window coordinate it is used together with
	         *		the selection options to carry out the selection operation by calling SelectByPoint */
            HPS.SelectionOptionsKit selectionOptions = new HPS.SelectionOptionsKit();
            HPS.SelectionResults selectionResults;
            HPS.Point selectionPoint;
	        ulong numberOfSelectedItems;
	        selectionOptions.SetLevel(HPS.Selection.Level.Entity).SetRelatedLimit(0).SetSorting(true);

	        view.Update();

            HPS.Key[] keyArray = new HPS.Key[9];
	        keyArray[0] = marker;
	        keyArray[1] = elements;
	        keyArray[2] = modelKey;
	        keyArray[3] = view.GetAttachedModelIncludeLink();
	        keyArray[4] = view.GetSegmentKey();
	        keyArray[5] = canvas.GetAttachedLayout().GetAttachedViewIncludeLink(0);
	        keyArray[6] = canvas.GetAttachedLayout().GetSegmentKey();
	        keyArray[7] = canvas.GetAttachedLayoutIncludeLink();
	        keyArray[8] = wk;

            HPS.KeyPath keyPath = new HPS.KeyPath();
	        keyPath.SetKeys(keyArray);
            keyPath.ConvertCoordinate(HPS.Coordinate.Space.World, new HPS.Point(-0.05f, -0.1f, 0), HPS.Coordinate.Space.Window, out selectionPoint);

	        numberOfSelectedItems = wk.GetSelectionControl().SelectByPoint(selectionPoint, selectionOptions, out selectionResults);

	        /* This block of code completes the following operations:
	         * 1. Obtains an iterator for the selection result list, and walks through it
	         * 2. For each selected item, it obtain its key, and then uses the key to obtain the
	         *		component selected, using the CADModel::GetComponentFromKey function
	         * 3. Obtains a Key Path for the selected component, and fully resolves it to the window
	         * 4. Highlights the selected component using its key path */
            HPS.SelectionResultsIterator it = selectionResults.GetIterator();
            HPS.HighlightOptionsKit hok = new HPS.HighlightOptionsKit();
	        hok.SetStyleName("orangeHighlight");
	        while(it.IsValid())
	        {
                HPS.SelectionItem selected = it.GetItem();
                HPS.Key geometryKey;
		        selected.ShowSelectedItem(out geometryKey);
                HPS.Component selectedComponent = modelFile.GetComponentFromKey(geometryKey);
                HPS.KeyPath[] selectionKeyPath = HPS.Component.GetKeyPath(selectedComponent);

		        wk.GetHighlightControl().Highlight(selectionKeyPath[0], hok);
		        it.Next();
	        }

            description.InsertText(new HPS.Point(0, 0.90f, 0), "The component under the red cross");
            description.InsertText(new HPS.Point(0, 0.85f, 0), "is selected and highlighted in orange");

            view.Update(HPS.Window.UpdateControl.CompleteUpdate);

            }
            catch (IOException e)
            {
                System.Console.WriteLine(e.Message);
            }
           
        }
    }
}
